@@ -11,6 +11,8 @@ except ImportError:
     import json
 
 
+ACTIVITY_GUID = int(uuid.uuid4().hex[0:8], 16)  
+
 class Activity(object):
     """vipy.object.Activity class
     
@@ -26,9 +28,9 @@ class Activity(object):
     >>> a = vipy.object.Activity(startframe=0, endframe=10, category='Walking', tracks={t.id():t})
 
     """
-    def __init__(self, startframe, endframe, framerate=None, label=None, shortlabel=None, category=None, tracks=None, attributes=None, actorid=None):
+    def __init__(self, startframe, endframe, framerate=None, label=None, shortlabel=None, category=None, tracks=None, attributes=None, actorid=None, confidence=None):
         assert not (label is not None and category is not None), "Activity() Constructor requires either label or category kwargs, not both"
-        assert startframe < endframe, "Start frame must be strictly less than end frame"
+        assert startframe <= endframe, "Start frame must be less than or equal to end frame"
         if tracks:
             tracks = [v for (k,v) in tracks.items()] if isinstance(tracks, dict) else tracks  # backwards compatible, track dictionary input
             assert (all([isstring(t) for t in tracks]) or all([isinstance(t, Track) for t in tracks])), "Invalid track input"
@@ -42,7 +44,7 @@ class Activity(object):
             if actorid not in trackid:
                 trackid.add(actorid)
 
-        self._id = uuid.uuid4().hex
+        global ACTIVITY_GUID; self._id = hex(int(ACTIVITY_GUID))[2:];  ACTIVITY_GUID = ACTIVITY_GUID + 1;  # faster, increment package level UUID4 initialized GUID
         self._startframe = int(startframe)
         self._endframe = int(endframe)
         self._framerate = framerate
@@ -51,13 +53,25 @@ class Activity(object):
         self._trackid = trackid
         self._actorid = actorid
 
-        self.attributes = attributes if attributes is not None else {}            
+        self.attributes = copy.copy(attributes) if attributes is not None else {}  # shallow copy
+        if confidence is not None:
+            self.attributes['confidence'] = float(confidence)
 
+    def hasattribute(self, k):
+        return k in self.attributes
+
+    def confidence(self, c=None):
+        if c is not None:
+            self.attributes['confidence'] = float(c)
+            return self
+        else:
+            return self.attributes['confidence'] if 'confidence' in self.attributes else None
+    
     @classmethod
     def from_json(obj, s):
         d = json.loads(s) if not isinstance(s, dict) else s                
-        return obj(startframe=d['_startframe'],
-                   endframe=d['_endframe'],
+        return obj(startframe=int(d['_startframe']),
+                   endframe=int(d['_endframe']),
                    framerate=d['_framerate'],
                    category=d['_label'],
                    shortlabel=d['_shortlabel'],
@@ -70,14 +84,14 @@ class Activity(object):
         return max(0, self.endframe() - self.startframe())
 
     def __repr__(self):
-        return str('<vipy.activity: category="%s", frames=(%d,%d), tracks=%s>' % (self.category(), self.startframe(), self.endframe(), len(self.trackids())))
+        return str('<vipy.activity: category="%s", frames=(%d,%d), tracks=%s%s>' % (self.category(), self.startframe(), self.endframe(), len(self.trackids()), '' if self.confidence() is None else ', confidence=%1.2f' % self.confidence()))
 
     def dict(self):
         """Return a python dictionary containing the relevant serialized attributes suitable for JSON encoding"""
         return self.json(s=None, encode=False)
 
     def json(self, encode=True):
-        d = {k:v if k != '_trackid' else list(v) for (k,v) in self.__dict__.items()}
+        d = {k:v if k != '_trackid' else tuple(v) for (k,v) in self.__dict__.items()}
         return json.dumps(d) if encode else d
     
     def actorid(self, actorid=None):
@@ -104,12 +118,18 @@ class Activity(object):
     def middleframe(self):
         return int(np.round((self.endframe() - self.startframe()) / 2.0)) + self.startframe()
 
-    def framerate(self, fps=None):
+    def framerate(self, fps=None, speed=None):
         """Resample (startframe, endframe) from known original framerate set by constructor to be new framerate fps"""        
-        if fps is None:
+        if fps is None and speed is None:
             return self._framerate
         else:
+            assert fps is not None or speed is not None, "Invalid input"
+            assert not (fps is not None and speed is not None), "Invalid input"
+            assert speed is None or speed > 0, "Invalid speed, must specify speed multiplier s=1, s=2 for 2x faster, s=0.5 for half slower"                    
+            assert fps is None or fps > 0, "Invalid framerate"
             assert self._framerate is not None, "Framerate conversion requires that the framerate is known for current activities.  This must be provided to the vipy.object.Activity() constructor."
+
+            fps = fps if fps is not None else (1.0/speed)*self._framerate            
             (self._startframe, self._endframe) = [int(np.round(f*(fps/float(self._framerate)))) for f in (self._startframe, self._endframe)]
             self._framerate = fps
             return self
@@ -121,7 +141,7 @@ class Activity(object):
         else:
             return self._label
 
-    def label(self, label):
+    def label(self, label=None):
         """Alias for category"""
         return self.category(label)
 
@@ -148,8 +168,10 @@ class Activity(object):
         """Return a set of track IDs associated with this activity"""
         return set(self._trackid)
 
-    def hasoverlap(self, other):
-        return self.temporal_iou(other) > 0
+    def hasoverlap(self, other, threshold=0):
+        assert isinstance(other, Activity), "Invalid input"
+        assert threshold >= 0 and threshold <= 1, "Invalid temporal IOU threshold"
+        return self.temporal_iou(other) > threshold
         
     def isneighbor(self, other, framegate=10):
         return self.temporal_iou(other.clone().temporalpad(framegate)) > 0 
@@ -160,21 +182,37 @@ class Activity(object):
         trackid = track.id() if isinstance(track, Track) else track
         return trackid in self._trackid
 
+    def append(self, newtrack):
+        """Append newtrack to this activity and set as actorid()"""
+        assert isinstance(newtrack, Track), "Invalid input - must be vipy.object.Track"
+        self._trackid.add(newtrack.id())
+        self.actorid(newtrack.id())
+        return self
+
+    def trackfilter(self, f):
+        """Remove all tracks such that the lambda function f(trackid) resolves to False"""
+        self._trackid = [tid for tid in self._trackid if f(tid)]
+        if not f(self.actorid()):
+            self._actorid = None
+        return self
+
     def replace(self, oldtrack, newtrack):
         """Replace oldtrack with newtrack if present in self._tracks.  Pass in a trackdict to share reference to track, so that track owner can modify the track and this object observes the change"""
         assert isinstance(oldtrack, Track) and isinstance(newtrack, Track), "Invalid input - must be vipy.object.Track"
-        self._trackid.discard(oldtrack.id())
-        self._trackid.add(newtrack.id())
-        if self.actorid() == oldtrack.id():
-            self.actorid(newtrack.id())
+        if self.hastrack(oldtrack):
+            self._trackid.discard(oldtrack.id())
+            self._trackid.add(newtrack.id())
+            if self.actorid() == oldtrack.id():
+                self.actorid(newtrack.id())
         return self    
 
     def replaceid(self, oldtrackid, newtrackid):
         """Replace oldtrack with newtrack if present in self._tracks.  Pass in a trackdict to share reference to track, so that track owner can modify the track and this object observes the change"""
-        self._trackid.discard(oldtrackid)
-        self._trackid.add(newtrackid)
-        if self.actorid() == oldtrackid:
-            self.actorid(newtrackid)
+        if self.hastrack(oldtrackid):
+            self._trackid.discard(oldtrackid)
+            self._trackid.add(newtrackid)
+            if self.actorid() == oldtrackid:
+                self.actorid(newtrackid)
         return self    
     
     def during(self, frame):
@@ -182,12 +220,31 @@ class Activity(object):
         return int(frame) >= self._startframe and int(frame) <= self._endframe
 
     def during_interval(self, startframe, endframe):
-        """Is the activity occurring for all frames within the interval [startframe, endframe) (non-inclusive of endframe)?"""
-        return all([self.during(f) for f in range(startframe, endframe)])
-    
+        """Is the activity occurring for any frames within the interval [startframe, endframe) (non-inclusive of endframe)?"""
+        return any([self.during(f) for f in range(startframe, endframe)]) if startframe <= endframe else False
+
+    def union(self, other, confweight=0.5):
+        """Compute the union of the new activity other to this activity by updating the start and end times and computing the mean confidence.
+        
+           -Note: other must have the same category and track IDs as self
+           -confweight [0,1]:  the convex combinatiopn weight applied to the new activity 
+        """
+        assert isinstance(other, Activity), "Invalid input"
+        assert self.actorid() == other.actorid(), "Actor ID must be the same"
+        assert self.category() == other.category(), "Assigned activity is a different category"
+        assert self.framerate() == other.framerate(), "Invalid input"
+        assert confweight >= 0 and confweight <= 1, "Confidence weight must be [0,1]"
+
+        self.startframe(min(other.startframe(), self.startframe()))
+        self.endframe(max(other.endframe(), self.endframe()))
+        if other.confidence() is not None and self.confidence() is not None:
+            self.confidence(float((1.0-confweight)*self.confidence() + confweight*other.confidence()))   # mean confidence
+        return self
+        
     def temporal_iou(self, other):
         """Return the temporal intersection over union of two activities"""
         assert isinstance(other, Activity), "Invalid input - must be vipy.object.Activity()"
+        assert self.framerate() == other.framerate(), "invald input - framerate must match"
         t_start = min(self.startframe(), other.startframe())
         t_end = max(self.endframe(), other.endframe())
         t_union = float(t_end - t_start)
@@ -204,11 +261,11 @@ class Activity(object):
         return self
 
     def truncate(self, startframe=None, endframe=None):
+        """Truncate the activity so that it is between startframe and endframe"""
         self._startframe = self._startframe if startframe is None else max(self._startframe, startframe)
-        self._endframe = self._endframe if endframe is None else max(self._endframe, endframe)
+        self._endframe = self._endframe if endframe is None else min(self._endframe, endframe)
         return self
-        
-    
+            
     def id(self, newid=None):
         if newid is None:
             return self._id
@@ -216,8 +273,12 @@ class Activity(object):
             self._id = newid
             return self
 
-    def clone(self):
-        return copy.deepcopy(self)
+    def clone(self, rekey=False):
+        #a = copy.deepcopy(self)
+        a = Activity.from_json(self.json(encode=False))
+        if rekey:
+            global ACTIVITY_GUID; a.id(newid=hex(int(ACTIVITY_GUID))[2:]);  ACTIVITY_GUID = ACTIVITY_GUID + 1;  # faster, increment package level UUID4 initialized GUID
+        return a
     
     def temporalpad(self, df):
         """Add a temporal pad of df=(before,after) or df=pad frames to the start and end of the activity.  The padded start frame may be negative."""
@@ -226,13 +287,17 @@ class Activity(object):
         self._endframe += int(df[1])
         return self  
 
+    def padto(self, t):
+        """Add a symmetric temporal pad so that the activity is at least t seconds long"""
+        return self.temporalpad((t - len(self))/2.0) if t > len(self) else self
+
     def disjoint(self, other, strict=False):
         """Enforce disjoint activities with other by shifting the endframe or startframe of self to not overlap if they share the same tracks.
            Other may be an Activity() or list of Activity()
            if strict=True, then throw an exception if other or self is fully contained with the other, resulting in degenerate activity after disjoint
         """
         for o in tolist(other):
-            assert isinstance(o, Activity), "Invalid input - must be vipy.object.Activity() or list of activities"       
+            assert isinstance(o, Activity), "Invalid input - must be vipy.activity.Activity() or list of activities"       
             if strict:
                 assert not (o.during(self.startframe()) and o.during(self.endframe())), "Self cannot fully overlap other"
                 assert not (self.during(o.startframe()) and self.during(o.endframe())), "Other cannot fully overlap self"
@@ -242,3 +307,7 @@ class Activity(object):
                 self.startframe(o.endframe()+1)
         return self  # may be zero length now
 
+    def temporal_distance(self, other):
+        """Return the temporal distance in frames between self and other which is the minimum frame difference between the end of one to the start of the other, or zero if they overlap"""
+        assert isinstance(other, Activity), "Invalid input - must be vipy.activity.Activity()"
+        return (max(self.startframe(), other.startframe()) - min(self.endframe(), other.endframe())) if self.temporal_iou(other) == 0 else 0
