@@ -9,7 +9,6 @@ import tempfile
 import time
 from time import gmtime, strftime, localtime
 from datetime import datetime
-import pytz
 import sys
 import csv
 import hashlib
@@ -28,6 +27,7 @@ import pathlib
 import socket
 import warnings
 import copy
+import bz2
 
 try:
     import ujson as json  # faster
@@ -35,6 +35,20 @@ except ImportError:
     import json
 
 
+def bz2pkl(filename, obj=None):
+    """Read/Write compressed pickle file"""
+    assert filename[-8:] == '.pkl.bz2', "Invalid filename - must be '*.pkl.bz2'"
+    if obj is not None:
+        f = bz2.BZ2File(filename, 'wb')
+        cPickle.dump(obj, f)
+        f.close()
+        return filename
+    else:
+        f = bz2.BZ2File(filename, 'rb')
+        obj = cPickle.load(f)
+        f.close()
+        return obj
+        
 
 def mergedict(d1, d2):
     assert isinstance(d1, dict) and isinstance(d2, dict)
@@ -134,7 +148,7 @@ def keymin(d):
 
 
 def isjsonfile(filename):
-    return hasextension(filename) and fileext(filename).lower() == '.json'
+    return len(filename) > 5 and filename[-5:] == '.json'
 
 
 def writejson(d, outfile):
@@ -167,6 +181,10 @@ def countby(inset, keyfunc):
     """Return dictionary of keys and group sizes for a grouping of the input list by keyfunc lambda function""" 
     return {k:len(v) for (k,v) in groupbyasdict(inset, keyfunc).items()}
 
+def most_frequent(inset):
+    """Return the most frequent element as determined by element equality"""
+    return sorted([(k,v) for (k,v) in vipy.util.countby(inset, lambda x: x).items()], key=lambda y: y[1])[-1][0]
+
 def countbyasdict(inset, keyfunc):
     """Alias for countby"""
     return countby(inset, keyfunc)
@@ -211,7 +229,7 @@ def dividelist(inlist, fractions):
     assert all([f >= 0 and f <=1 for f in fractions])
     assert np.sum(fractions) == 1
     assert len(inlist) >= len(fractions)
-    N = np.int32(np.maximum(1, np.ceil(len(inlist)*np.array(fractions))))
+    N = np.int32(np.maximum(0, np.ceil(len(inlist)*np.array(fractions))))
     outlist = []
     for n in N:
         outlist.append(inlist[0:n])
@@ -239,10 +257,8 @@ def chunklistWithOverlap(inlist, size_per_chunk, overlap_per_chunk):
     """Convert list into a list of lists such that each element is a list
     containing a sequential chunk of the original list of length
     size_per_chunk"""
-    assert size_per_chunk >= 1 and overlap_per_chunk >= 0 and \
-        size_per_chunk > overlap_per_chunk
-    return [inlist[max(0, i-int(np.floor(size_per_chunk/2))):i+int(np.ceil(size_per_chunk/2))] for i in range(
-        0, len(inlist), size_per_chunk - overlap_per_chunk)]
+    assert size_per_chunk >= 1 and overlap_per_chunk >= 0 and size_per_chunk > overlap_per_chunk
+    return [inlist[i-size_per_chunk:i] for i in range(size_per_chunk, len(inlist)+(size_per_chunk-overlap_per_chunk), size_per_chunk - overlap_per_chunk)]
 
 
 def chunklistwithoverlap(inlist, size_per_chunk, overlap_per_chunk):
@@ -486,14 +502,16 @@ def loadas(infile, format='dill'):
         raise ValueError('unknown serialization format "%s"' % format)
 
 
-def load(infile):
+def load(infile, abspath=False):
     """Load variables from a relocatable archive file format, either Dill Pickle or JSON.
        
        Loading is performed by attemping the following:
 
        1. load the pickle or json file
-       2. If the loaded object is a vipy object (or iterable) and the relocatable path /$PATH is present, try to repath it to the directory containing this archive
-       3. If the resulting files are not found, throw a warning
+       2. if abspath=true, then convert relative paths to absolute paths for object when loaded
+       3. If the loaded object is a vipy object (or iterable) and the relocatable path /$PATH is present, try to repath it to the directory containing this archive (this has been deprecated)
+       4. If the resulting files are not found, throw a warning
+       5. If a large number of objects are loaded, disable garbage collection.
 
     """
     infile = os.path.abspath(os.path.expanduser(infile))
@@ -519,11 +537,35 @@ def load(infile):
             warnings.warn('Loading "%s" that contains redistributable paths - Use vipy.util.distload("%s", datapath="/path/to/your/data") to rehome absolute file paths' % (infile, infile))
     elif hasattr(testobj, 'filename') and testobj.filename() is not None and not os.path.exists(testobj.filename()):
         if hasattr(testobj, 'hasurl') and testobj.hasurl():
-            warnings.warn('Invalid file - Loading "%s" that contains filename "%s" which does not exist - This must be downloaded from the provided url() using download()' % (infile, testobj.filename()))
+            warnings.warn('Loading archive "%s" that contains filename "%s" which does not exist - This must be downloaded from the provided url() using download()' % (infile, testobj.filename()))
         elif not os.path.isabs(testobj.filename()):
-            warnings.warn('Invalid file - Loading "%s" that contains relative path "%s" which does not exist - Loading archives with relative paths must be loaded from the same directory containing the videos' % (infile, testobj.filename()))
+            if not abspath:
+                warnings.warn('Loading archive "%s" with relative paths.  Changing directory to "%s".  Disable this warning by setting vipy.util.load(..., abspath=True).' % (infile, filepath(infile)))
+                os.chdir(filepath(infile))
+            else:
+                # Absolute path?  The loaded archive will no longer be relocatable, and the videos directory cannot be moved
+                pwd = os.getcwd()  
+                os.chdir(filepath(infile))  # change to archive directory
+                objout = [o.abspath() if o.hasfilename() else o for o in tolist(obj)]  # set absolute paths
+                obj = objout if isinstance(obj, list) else objout[0]
+                if any([not o.hasfilename() for o in tolist(obj)]):
+                    warnings.warn('Loading "%s" that contains absolute path "%s" which does not exist' % (infile, tolist(obj)[0]))
+                os.chdir(pwd)  # set it back
         else:
-            warnings.warn('Invalid file - Loading "%s" that contains absolute path "%s" which does not exist' % (infile, testobj.filename()))
+            warnings.warn('Loading "%s" that contains absolute path "%s" which does not exist' % (infile, testobj.filename()))
+
+    # Large vipy object?  Disable garbage collection.
+    #   - Python uses reference counting for the primary garbage collection mechanism, but also uses reference cycle checks to search for dependencies between objects.
+    #   - All vipy objects are self contained, and do not have reference cycles.  However, there is no way to mark an individual object which does not participate in reference cycle counting.
+    #   - This means that a large number of vipy objects, garbage collection can take minutes searching for cycles which are never there.  To fix this, globally disable the garbage collector.
+    #   - Note that refernece counting is still performed, we are just disabling reference *cycle* counting using the generational garbage collector.
+    #   - This can be re-enabled at any time by "import gc; gc.enable()"
+    #   - If you use %autoreload iPython magic command, note that this will be very slow.  You should set %sutoreload 0
+    #   - Alternatively, load as JSON and all attributes will be unpacked on demand and stored in a packed format that is not tracked (e.g. tuple of strings) by the reference cycle counter
+    if format != 'json' and hasattr(testobj, 'filename') and len(tolist(obj)) > 10000:
+        # Should we warn the user?  Probably doesn't matter ...
+        #warnings.warn('Loading "%s" that contains a large number of vipy objects - disabling reference cycle checks.  Re-enable at any time using "import gc; gc.enable()"' % (infile))
+        import gc; gc.disable()
     return obj
 
 
@@ -574,13 +616,13 @@ def repath(v, srcpath, dstpath):
 def scpsave(V):
     import vipy.image
     import vipy.video
-    if (isinstance(V, vipy.image.Image) or isinstance(V, vipy.video.Video)) and v.hasfilename():        
+    if (isinstance(V, vipy.image.Image) or isinstance(V, vipy.video.Video)) and V.hasfilename():        
         v = V
         v = v.clone().url('scp://%s:%s' % (socket.gethostname(), v.filename())).nofilename()
     elif islist(V) and all([isinstance(v, vipy.image.Image) or isinstance(v, vipy.video.Video) for v in V]):
         v = [v.clone().url('scp://%s:%s' % (socket.gethostname(), v.abspath().filename())).nofilename() for v in V]
     else:
-        pass  # no vipy objects
+        v = V # no vipy objects
 
     pklfile = 'scp://%s:%s' % (socket.gethostname(), save(v, temppkl()))
     cmd = "V = vipy.util.scpload('%s')" % pklfile
@@ -691,7 +733,7 @@ def ishtml(filename):
 
 def ispickle(filename):
     """Is the file a pickle archive file"""
-    return isfile(filename) and os.path.exists(filename) and (fileext(filename) is not None) and fileext(filename).lower() in ['.pk', '.pkl']
+    return isfile(filename) and os.path.exists(filename) and (((fileext(filename) is not None) and fileext(filename).lower() in ['.pk', '.pkl']) or (filename[-4:] == '.pkl'))
 
 
 def ishdf5(path):
@@ -732,6 +774,28 @@ def newpath(filename, newdir):
     (head, tail) = os.path.split(filename)
     return os.path.join(newdir, tail)
 
+def newprefix(filename, newprefix, depth=0):
+    """Return /a/b/c/h/i.ext for filename /f/g/h/i.ext and prefix /a/b/c and depth=1"""
+    p = filepath(filename, depth=depth)
+    return os.path.normpath(filename.replace(p, newprefix))
+
+def newpathdir(filename, olddir, newdir, n=1):
+    """Return /a/b/n/d/e.ext for filename=/a/b/c/d/e.ext, olddir=c, newdir=n"""
+    p = pathlib.PurePath(filename)
+    assert sum([d == olddir for d in p.parts]) == n, "Path must have exactly %s directory matches" % n
+    return os.path.join(*[d.replace(olddir, newdir) for d in list(p.parts)])
+
+
+def newpathroot(filename, newroot):
+    """Return /r/b/c.ext for filename /a/b/c.ext and new root directory r"""
+    p = pathlib.PurePath(filename)
+    path = list(p.parts)    
+    if len(p.root) == 0:
+        path[0] = newroot
+    else:
+        path[1] = newroot
+    return os.path.join(*path)
+
 def topath(filename, newdir):
     """Alias for newpath"""
     return newpath(filename, newdir)
@@ -739,7 +803,7 @@ def topath(filename, newdir):
 
 def filefull(f):
     """Return /a/b/c for filename /a/b/c.ext"""
-    return splitextension(f)[0]
+    return f.replace(fileext(f, multidot=True, withdot=True), '')  # better handling of .tar.gz
 
 
 def filetail(filename):
@@ -797,8 +861,14 @@ def listext(indir, ext):
 
 def jsonlist(indir):
     """return list of fJSON iles with absolute path in a directory"""
-    return extlist(indir, ext='.json')
+    #return extlist(indir, ext='.json')  # FIXME: broken.for.wonky.filenames.with.dots.json
+    return [os.path.abspath(os.path.join(indir, item))
+            for item in os.listdir(indir)
+            if len(item) > 5 and item[-5:] == '.json']
 
+def listjson(indir):
+    """Alias for jsonlist"""
+    return jsonlist(indir)
 
 def writelist(mylist, outfile, mode='w'):
     """Write list of strings to an output file with each row an element of
@@ -1309,8 +1379,19 @@ class Timer(object):
        >>> [some more code]
        >>> print(t)
 
+       >>> with Timer():
+       >>>    [some code]
+       
+
     """
-    def __init__(self, sprintf='timer: elapsed=%1.6fs, since=%1.6fs'):
+    def __enter__(self):
+        self._begin = time.time()
+        self._last = self._begin
+
+    def __exit__(self, *args):
+        print(self.__repr__())
+
+    def __init__(self, sprintf='[vipy.util.timer]: elapsed=%1.6fs, since=%1.6fs'):
         self._sprintf = sprintf
         self._begin = time.time()
         self._last = self._begin
@@ -1340,8 +1421,8 @@ def timestamp():
     return str.upper(strftime("%d%b%y_%I%M%S%p", localtime()))
 
 def clockstamp():
-    """Datetime stamp in eastern timezone with second resolution"""    
-    return datetime.now().astimezone(pytz.timezone("US/Eastern")).strftime("%Y-%m-%dT%H:%M:%S%z")    
+    """Datetime stamp in local timezone with second resolution with format Year-Month-Day Hour:Minute:Second"""    
+    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")    
 
 def minutestamp():
     """Return date and time string in form DDMMMYY_HHMM"""
@@ -1374,6 +1455,10 @@ def premkdir(filename):
     return remkdir(filepath(filename))
 
 
+def newbase(filename, base):
+    """Convert filename=/a/b/c.ext base=d -> /a/b/d.ext"""
+    return os.path.join(filepath(filename), '%s.%s' % (base, fileext(filename, withdot=False)))
+
 def toextension(filename, newext):
     """Convert filenam='/path/to/myfile.ext' to /path/to/myfile.xyz, such that newext='xyz' or newext='.xyz'"""
     if '.' in newext:
@@ -1386,7 +1471,7 @@ def splitextension(filename):
     """Given /a/b/c.ext return tuple of strings ('/a/b/c', '.ext')"""
     (head, tail) = os.path.split(filename)
     try:
-        (base, ext) = str.split(tail, '.', 1)  # for .tar.gz
+        (base, ext) = str.rsplit(tail, '.', 1)  # for .tar.gz
         ext = '.' + ext
     except:
         base = tail
@@ -1399,8 +1484,8 @@ def hasextension(filename):
     return fileext(filename) is not None
 
 
-def fileext(filename, multidot=True):
-    """Given filename /a/b/c.ext return '.ext', or /a/b/c.tar.gz return '.tar.gz'.   If multidot=False, then return '.gz'"""
+def fileext(filename, multidot=True, withdot=True):
+    """Given filename /a/b/c.ext return '.ext', or /a/b/c.tar.gz return '.tar.gz'.   If multidot=False, then return '.gz'.  If withdot=False, return 'ext'"""
     (head, tail) = os.path.split(filename)
     try:
         parts = str.rsplit(tail, '.', 2)
@@ -1414,7 +1499,7 @@ def fileext(filename, multidot=True):
     except:
         base = tail
         ext = None
-    return ext
+    return ext if withdot else ext[1:]
 
 def mediaextension(filename):
     """Return '.mp4' for filename='/a/b/c.mp4'"""
